--
-- sprayerEffect-Script
-- 
-- Author: Dominik Raith (DR_Modding)
--
-- default sprayer specialization but:
		-- no turnOff if Empty 
		-- only working if given fillUnit exists.
		-- only show effects if (pickup) is lowered
		-- no need to be filled to turn on the vehicle (foragewagon)
--
--
------------------------------------------------------------------------------------------------------
-- add new Specialization: sprayerEffect

sprayerEffect = {};
sprayerEffect.MOD_NAME = g_currentModName;







function sprayerEffect.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", sprayerEffect)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdate", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onPreDetach", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onStartWorkAreaProcessing", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onEndWorkAreaProcessing", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onStateChange", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onSetLowered", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onFillUnitFillLevelChanged", sprayerEffect)
    SpecializationUtil.registerEventListener(vehicleType, "onSprayTypeChange", sprayerEffect)
    --SpecializationUtil.registerEventListener(vehicleType, "onAIImplementEnd", sprayerEffect)
end


function sprayerEffect.registerEvents(vehicleType)
    SpecializationUtil.registerEvent(vehicleType, "onSprayTypeChange")
end


function sprayerEffect.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "processSprayerArea",                            sprayerEffect.processSprayerArea)
    SpecializationUtil.registerFunction(vehicleType, "getExternalFill",                               sprayerEffect.getExternalFill)
    SpecializationUtil.registerFunction(vehicleType, "getAreEffectsVisible",                          sprayerEffect.getAreEffectsVisible)
    SpecializationUtil.registerFunction(vehicleType, "getSprayerUsage",                               sprayerEffect.getSprayerUsage)
    --SpecializationUtil.registerFunction(vehicleType, "getUseSprayerAIRequirements",                   sprayerEffect.getUseSprayerAIRequirements)
    --SpecializationUtil.registerFunction(vehicleType, "setSprayerAITerrainDetailProhibitedRange",      sprayerEffect.setSprayerAITerrainDetailProhibitedRange)
    SpecializationUtil.registerFunction(vehicleType, "getSprayerFillUnitIndex",                       sprayerEffect.getSprayerFillUnitIndex)
    SpecializationUtil.registerFunction(vehicleType, "loadSprayTypeFromXML",                          sprayerEffect.loadSprayTypeFromXML)
    SpecializationUtil.registerFunction(vehicleType, "getActiveSprayType",                            sprayerEffect.getActiveSprayType)
    SpecializationUtil.registerFunction(vehicleType, "getIsSprayTypeActive",                          sprayerEffect.getIsSprayTypeActive)
end


function sprayerEffect.registerOverwrittenFunctions(vehicleType)
   -- SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAIMarkers",                    sprayerEffect.getAIMarkers)
   -- SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAISizeMarkers",                sprayerEffect.getAISizeMarkers)
   -- SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAIImplementCollisionTriggers", sprayerEffect.getAIImplementCollisionTriggers)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDrawFirstFillText",            sprayerEffect.getDrawFirstFillText)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanToggleTurnedOn",            sprayerEffect.getCanToggleTurnedOn)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeTurnedOn",                sprayerEffect.getCanBeTurnedOn)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "loadWorkAreaFromXML",             sprayerEffect.loadWorkAreaFromXML)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getIsWorkAreaActive",             sprayerEffect.getIsWorkAreaActive)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "doCheckSpeedLimit",               sprayerEffect.doCheckSpeedLimit)
   -- SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAIRequiresTurnOffOnHeadland",  sprayerEffect.getAIRequiresTurnOffOnHeadland)
  --  SpecializationUtil.registerOverwrittenFunction(vehicleType, "getDirtMultiplier",               sprayerEffect.getDirtMultiplier)
  --  SpecializationUtil.registerOverwrittenFunction(vehicleType, "getWearMultiplier",               sprayerEffect.getWearMultiplier)
end







function sprayerEffect:doCheckSpeedLimit(superFunc)
    return superFunc(self) or (self:getIsTurnedOn() and self.spec_sprayerEffect.useSpeedLimit)
end


function sprayerEffect:getActiveSprayType()
    local spec = self.spec_sprayerEffect
    for _, sprayType in ipairs(spec.sprayTypes) do
        if self:getIsSprayTypeActive(sprayType) then
            return sprayType
        end
    end
    return nil
end


function sprayerEffect:getAreEffectsVisible()
    return true
end


function sprayerEffect:getCanBeTurnedOn(superFunc)
    local spec = self.spec_sprayerEffect
    if not spec.allowsSpraying then
        return false
    end
    if self:getFillUnitFillLevel(self:getSprayerFillUnitIndex()) <= 0 and spec.needsToBeFilledToTurnOn then
            local sprayVehicle = nil
            for _, supportedSprayType in ipairs(spec.supportedSprayTypes) do
                for _, src in ipairs(spec.fillTypeSources[supportedSprayType]) do
                    local vehicle = src.vehicle
                    if vehicle:getFillUnitFillType(src.fillUnitIndex) == supportedSprayType and vehicle:getFillUnitFillLevel(src.fillUnitIndex) > 0 then
                        sprayVehicle = vehicle
                        break
                    end
                end
            end
            if sprayVehicle == nil then
                return false
            end
    end
    return superFunc(self)
end



function sprayerEffect:getCanToggleTurnedOn(superFunc)
    if self.isClient then
        local spec = self.spec_sprayerEffect
        if spec.needsToBeFilledToTurnOn then
            if not self:getCanBeTurnedOn() and self:getFillUnitCapacity(self:getSprayerFillUnitIndex()) <= 0 then
                return false
            end
        end
    end
    return superFunc(self)
end


function sprayerEffect.getDefaultSpeedLimit()
    return 15
end


function sprayerEffect:getDirtMultiplier(superFunc)
    local spec = self.spec_sprayerEffect
    if spec.isWorking then
        return superFunc(self) + self:getWorkDirtMultiplier() * self:getLastSpeed() / self.speedLimit
    end
    return superFunc(self)
end


function sprayerEffect:getDrawFirstFillText(superFunc)
    if self.isClient then
        local spec = self.spec_sprayerEffect
        if spec.needsToBeFilledToTurnOn then
            if self:getIsActiveForInput() and self:getIsSelected() and not self.isAlwaysTurnedOn then
                if not self:getCanBeTurnedOn() and self:getFillUnitFillLevel(self:getSprayerFillUnitIndex()) <= 0 and self:getFillUnitCapacity(self:getSprayerFillUnitIndex()) > 0 then
                    return true
                end
            end
        end
    end
    return superFunc(self)
end



function sprayerEffect:getExternalFill(fillType, dt)
    local found = false
    local isUnknownFillType = fillType == FillType.UNKNOWN
    local allowLiquidManure = self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.LIQUIDMANURE)
    local allowDigestate = self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.DIGESTATE)
    local allowManure = self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.MANURE)
    local allowLiquidFertilizer = self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.LIQUIDFERTILIZER)
    local allowFertilizer = self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.FERTILIZER)
    local allowsLiquidManureDigistate = allowLiquidManure or allowDigestate
    local usage = 0
    local farmId = self:getActiveFarm()
    local stats = g_currentMission:farmStats(self:getLastTouchedFarmlandFarmId())
    if fillType == FillType.LIQUIDMANURE or fillType == FillType.DIGESTATE or (isUnknownFillType and allowsLiquidManureDigistate) then
        if g_currentMission.missionInfo.helperSlurrySource == 2 then -- buy manure
            found = true
            if g_currentMission.economyManager:getCostPerLiter(FillType.LIQUIDMANURE, false) then
                fillType = FillType.LIQUIDMANURE
            else
                fillType = FillType.DIGESTATE
            end
            usage = self:getSprayerUsage(fillType, dt)
            if self.isServer then
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType, false) * 1.5  -- increase price if AI is active to reward the player's manual work
                stats:updateStats("expenses", price)
                g_currentMission:addMoney(-price, farmId, MoneyType.PURCHASE_FERTILIZER)
            end
        elseif g_currentMission.missionInfo.helperSlurrySource > 2 then
            local info = g_currentMission.liquidManureTriggers[g_currentMission.missionInfo.helperSlurrySource-2]
            if info ~= nil then -- Can be nil if pen was removed
                local fillLevel = info.silo:getFillLevel(FillType.LIQUIDMANURE)
                if fillLevel > 0 then
                    found = true
                    usage = self:getSprayerUsage(FillType.LIQUIDMANURE, dt)
                    if self.isServer then
                        info.silo:setFillLevel(FillType.LIQUIDMANURE, fillLevel-usage)
                    end
                end
            end
        end    --
    elseif fillType == FillType.MANURE or (fillType == FillType.UNKNOWN and allowManure) then
        if fillType == FillType.UNKNOWN and allowManure then
            fillType = FillType.MANURE
        end
        if g_currentMission.missionInfo.helperManureSource == 2 then -- buy manure
            found = true
            fillType = FillType.MANURE
            usage = self:getSprayerUsage(fillType, dt)
            if self.isServer then
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType, false) * 1.5  -- increase price if AI is active to reward the player's manual work
                stats:updateStats("expenses", price)
                g_currentMission:addMoney(-price, farmId, MoneyType.PURCHASE_FERTILIZER)
            end
        elseif g_currentMission.missionInfo.helperManureSource > 2 then
            local info = g_currentMission.manureHeaps[g_currentMission.missionInfo.helperManureSource-2]
            if info ~= nil then -- Can be nil if pen was removed
                usage = self:getSprayerUsage(FillType.MANURE, dt)
                if self.isServer then
                    if info.manureHeap:removeManure(usage) > 0 then
                        found = true
                        fillType = FillType.MANURE
                    end
                end
            end
        end    --
    elseif fillType == FillType.FERTILIZER or
           fillType == FillType.LIQUIDFERTILIZER or
           fillType == FillType.HERBICIDE or
           fillType == FillType.LIME or
          (fillType == FillType.UNKNOWN and (allowLiquidFertilizer or allowFertilizer)) then
        if g_currentMission.missionInfo.helperBuyFertilizer then
            found = true
            if fillType == FillType.UNKNOWN then
                if self:getFillUnitAllowsFillType(self:getSprayerFillUnitIndex(), FillType.LIQUIDFERTILIZER) then
                    fillType = FillType.LIQUIDFERTILIZER
                else
                    fillType = FillType.FERTILIZER
                end
            end
            usage = self:getSprayerUsage(fillType, dt)
            if self.isServer then
                local price = usage * g_currentMission.economyManager:getCostPerLiter(fillType, false) * 1.5  -- increase price if AI is active to reward the player's manual work
                stats:updateStats("expenses", price)
                g_currentMission:addMoney(-price, farmId, MoneyType.PURCHASE_FERTILIZER)
            end
        end
    end
    if found then
        return fillType, usage
    end
    return FillType.UNKNOWN, 0
end





function sprayerEffect:getIsSprayTypeActive(sprayType)
    if sprayType.fillTypes ~= nil then
        local retValue = false
        local currentFillType = self:getFillUnitFillType(sprayType.fillUnitIndex or self.spec_sprayerEffect.fillUnitIndex)
        for _, fillType in ipairs(sprayType.fillTypes) do
            if currentFillType == g_fillTypeManager:getFillTypeIndexByName(fillType) then
                retValue = true
            end
        end
        if not retValue then
            return false
        end
    end
    return true
end


function sprayerEffect:getIsWorkAreaActive(superFunc, workArea)
    if workArea.sprayType ~= nil then
        local sprayType = self:getActiveSprayType()
        if sprayType ~= nil then
            if sprayType.index ~= workArea.sprayType then
                return false
            end
        end
    end
    return superFunc(self, workArea)
end

function sprayerEffect:getSprayerFillUnitIndex()
    local sprayType = self:getActiveSprayType()
    if sprayType ~= nil then
        return sprayType.fillUnitIndex
    end
    return self.spec_sprayerEffect.fillUnitIndex
end


function sprayerEffect:getSprayerUsage(fillType, dt)
    if fillType == FillType.UNKNOWN then
        return 0
    end
    local spec = self.spec_sprayerEffect
    local scale = Utils.getNoNil(spec.usageScale.fillTypeScales[fillType], spec.usageScale.default)
    local litersPerSecond = 1
    local sprayType = g_sprayTypeManager:getSprayTypeByFillTypeIndex(fillType)
    if sprayType ~= nil then
        litersPerSecond = sprayType.litersPerSecond
    end
    local activeSprayType = self:getActiveSprayType()
    local workWidth = spec.usageScale.workingWidth
    if activeSprayType ~= nil then
        workWidth = activeSprayType.usageScale.workingWidth or workWidth
    end
    return scale * litersPerSecond * self.speedLimit * workWidth * dt * 0.001
end




function sprayerEffect:getWearMultiplier(superFunc)
    local spec = self.spec_sprayerEffect
    if spec.isWorking then
        return superFunc(self) + self:getWorkWearMultiplier() * self:getLastSpeed() / self.speedLimit
    end
    return superFunc(self)
end


function sprayerEffect.initSpecialization()
    g_workAreaTypeManager:addWorkAreaType("sprayer", false)
end


function sprayerEffect:loadSprayTypeFromXML(xmlFile, key, sprayType)
    sprayType.fillUnitIndex = Utils.getNoNil(getXMLInt(xmlFile, key.. "#fillUnitIndex"), 1)
    sprayType.unloadInfoIndex = Utils.getNoNil(getXMLInt(xmlFile, key.. "#unloadInfoIndex"), 1)
    sprayType.fillVolumeIndex = getXMLInt(xmlFile, key.. "#fillVolumeIndex")
    sprayType.samples = {}
    sprayType.samples.work = g_soundManager:loadSampleFromXML(xmlFile, key.. ".sounds", "work", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
    sprayType.effects = g_effectManager:loadEffect(xmlFile, key.. ".effects", self.components, self, self.i3dMappings)
    sprayType.animationNodes = g_animationManager:loadAnimations(xmlFile, key.. ".animationNodes", self.components, self, self.i3dMappings)
    sprayType.turnedAnimation = getXMLString(self.xmlFile, key .. ".turnedAnimation#name") or ""
    local fillTypesStr = getXMLString(xmlFile, key.. "#fillTypes")
    if fillTypesStr ~= nil then
        sprayType.fillTypes = StringUtil.splitString(" ", fillTypesStr)
    end
    sprayType.objectChanges = {}
    ObjectChangeUtil.loadObjectChangeFromXML(self.xmlFile, key, sprayType.objectChanges, self.components, self)
    ObjectChangeUtil.setObjectChanges(sprayType.objectChanges, false)
    sprayType.usageScale = {}
    sprayType.usageScale.workingWidth = Utils.getNoNil(getXMLFloat(self.xmlFile, key..".usageScales#workingWidth"), 12)
    return true
end


function sprayerEffect:loadWorkAreaFromXML(superFunc, workArea, xmlFile, key)
    local retValue = superFunc(self, workArea, xmlFile, key)
    if workArea.type == WorkAreaType.DEFAULT then
        workArea.type = WorkAreaType.SPRAYER
    end
    workArea.sprayType = getXMLInt(xmlFile, key.. "#sprayType")
    return retValue
end



function sprayerEffect:onDelete()
    if self.isClient then
        local spec = self.spec_sprayerEffect
        g_effectManager:deleteEffects(spec.effects)
        for _,sample in pairs(spec.samples) do
            g_soundManager:deleteSample(sample)
        end
        g_animationManager:deleteAnimations(spec.animationNodes)
        for _, sprayType in ipairs(spec.sprayTypes) do
            g_effectManager:deleteEffects(sprayType.effects)
            g_soundManager:deleteSamples(sprayType.samples)
            g_animationManager:deleteAnimations(sprayType.animationNodes)
        end
    end
end



function sprayerEffect:onEndWorkAreaProcessing(dt, hasProcessed)
    local spec = self.spec_sprayerEffect
    if self.isServer then
        if spec.workAreaParameters.isActive then
            local sprayVehicle = spec.workAreaParameters.sprayVehicle
            local usage = spec.workAreaParameters.usage
            if sprayVehicle ~= nil then
                local sprayVehicleFillUnitIndex = spec.workAreaParameters.sprayVehicleFillUnitIndex
                local sprayFillType = spec.workAreaParameters.sprayFillType
                local unloadInfoIndex = spec.unloadInfoIndex
                local sprayType = self:getActiveSprayType()
                if sprayType ~= nil then
                    unloadInfoIndex = sprayType.unloadInfoIndex
                end
                local unloadInfo = self:getFillVolumeUnloadInfo(unloadInfoIndex)
                sprayVehicle:addFillUnitFillLevel(self:getOwnerFarmId(), sprayVehicleFillUnitIndex, -usage, sprayFillType, ToolType.UNDEFINED, unloadInfo)
            end
            local ha = MathUtil.areaToHa(spec.workAreaParameters.lastStatsArea, g_currentMission:getFruitPixelsToSqm())
            local stats = g_currentMission:farmStats(self:getLastTouchedFarmlandFarmId())
            stats:updateStats("workedHectares", ha)
            stats:updateStats("fertilizedHectares", ha)
            stats:updateStats("fertilizedTime", dt/(1000*60))
            stats:updateStats("sprayUsage", usage)
        end
    end
end


function sprayerEffect:onFillUnitFillLevelChanged(fillUnitIndex, fillLevelDelta, fillType, toolType, fillPositionData, appliedDelta)
    local fillLevel = self:getFillUnitFillLevel(fillUnitIndex)
    --[[
	if fillLevel == 0 and self:getIsTurnedOn() and not self:getIsAIActive() then
        local spec = self.spec_sprayerEffect
        -- only deactivate sprayer if no fill type source for the same fill type is available
        local hasValidSource = false
        if spec.fillTypeSources[fillType] ~= nil then
            for _, src in ipairs(spec.fillTypeSources[fillType]) do
                local vehicle = src.vehicle
                if vehicle:getIsFillUnitActive(src.fillUnitIndex) then
                    local vehicleFillType = vehicle:getFillUnitFillType(src.fillUnitIndex)
                    local vehicleFillLevel = vehicle:getFillUnitFillLevel(src.fillUnitIndex)
                    if vehicleFillLevel > 0 and vehicleFillType == fillType then
                        hasValidSource = true
                    end
                end
            end
        end
        if not hasValidSource then
            self:setIsTurnedOn(false)
        end
    end
	]]
	local spec = self.spec_sprayerEffect
    if self.isClient then
		if fillLevel == 0 then
			g_effectManager:stopEffects(spec.effects)
			g_animationManager:stopAnimations(spec.animationNodes)
			if spec.animationName ~= "" and self.stopAnimation ~= nil then
				self:stopAnimation(spec.animationName, true)
			end
			g_soundManager:stopSample(spec.samples.work)
			-- deactivate effects on all spray types (the spray type has may changed during activation)
			for _, sprayType in ipairs(spec.sprayTypes) do
				g_effectManager:stopEffects(sprayType.effects)
				g_animationManager:stopAnimations(sprayType.animationNodes)
				g_soundManager:stopSample(sprayType.samples.work)
				self:playAnimation(sprayType.turnedAnimation, -1, self:getAnimationTime(sprayType.turnedAnimation), true)
			end
		end
    end

	if self:getIsTurnedOn() == true then
		if self:getFillUnitFillLevel(spec.fillUnitIndex) > 0 and spec.sprayEffectAllowed == true then													
			if self:getAreEffectsVisible() then
				local fillType = self:getFillUnitLastValidFillType(self:getSprayerFillUnitIndex())
				if fillType == FillType.UNKNOWN then
					fillType = self:getFillUnitFirstSupportedFillType(self:getSprayerFillUnitIndex())
				end
				g_effectManager:setFillType(spec.effects, fillType)
				g_effectManager:startEffects(spec.effects)
				if sprayType ~= nil then
					g_effectManager:setFillType(sprayType.effects, fillType)
					g_effectManager:startEffects(sprayType.effects)
					g_animationManager:startAnimations(sprayType.animationNodes)
				end
				g_animationManager:startAnimations(spec.animationNodes)
			end
		end
	end
	
	
end



function sprayerEffect:onLoad(savegame)
    self.spec_sprayerEffect = self[string.format("spec_%s.sprayerEffect", sprayerEffect.MOD_NAME)]
	local spec = self.spec_sprayerEffect
    XMLUtil.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.sprayParticles.emitterShape", "vehicle.sprayer.effects.effectNode#effectClass='ParticleEffect'") --FS17 to FS19
    spec.allowsSpraying = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#allowsSpraying"), true)
    spec.needsTankActivation = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#needsTankActivation"), false)
    spec.activateTankOnLowering = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#activateTankOnLowering"), false)
    spec.activateOnLowering = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.sprayer#activateOnLowering"), false)
    spec.usageScale = {}
    spec.usageScale.default = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.sprayer.usageScales#scale"), 1)
    spec.usageScale.workingWidth = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.sprayer.usageScales#workingWidth"), 12)
    spec.usageScale.fillTypeScales = {}
    local i=0
    while true do
        local key = string.format("vehicle.sprayer.usageScales.sprayUsageScale(%d)", i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end
        local fillTypeStr = getXMLString(self.xmlFile, key.. "#fillType")
        local scale = getXMLFloat(self.xmlFile, key.. "#scale")
        if fillTypeStr ~= nil and scale ~= nil then
            local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(fillTypeStr)
            if fillTypeIndex ~= nil then
                spec.usageScale.fillTypeScales[fillTypeIndex] = scale
            else
                print("Warning: Invalid spray usage scale fill type '"..fillTypeStr.."' in '" .. self.configFileName.. "'")
            end
        end
        i = i+1
    end
    spec.sprayTypes = {}
    i = 0
    while true do
        local key = string.format("vehicle.sprayer.sprayTypes.sprayType(%d)", i)
        if not hasXMLProperty(self.xmlFile, key) then
            break
        end
        local sprayType = {}
        if self:loadSprayTypeFromXML(self.xmlFile, key, sprayType) then
            table.insert(spec.sprayTypes, sprayType)
            sprayType.index = #spec.sprayTypes
        end
        i = i + 1
    end
    spec.lastActiveSprayType = nil
    if self.isClient then
        spec.effects = g_effectManager:loadEffect(self.xmlFile, "vehicle.sprayer.effects", self.components, self, self.i3dMappings)
        spec.animationName = Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.sprayer.animation#name"), "")
        spec.samples = {}
        spec.samples.work = g_soundManager:loadSampleFromXML(self.xmlFile, "vehicle.sprayer.sounds", "work", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
        spec.sampleFillEnabled = false
        spec.sampleFillStopTime = -1
        spec.lastFillLevel = -1
        spec.animationNodes = g_animationManager:loadAnimations(self.xmlFile, "vehicle.sprayer.animationNodes", self.components, self, self.i3dMappings)
    end
    if self.sowingMachineGroundContactFlag == nil and self.cultivatorGroundContactFlag == nil then
        --if self.addAITerrainDetailProhibitedRange ~= nil then
        --    self:addAITerrainDetailProhibitedRange(1, 2, g_currentMission.sprayFirstChannel, g_currentMission.sprayNumChannels)
        --    self:addAITerrainDetailProhibitedRange(g_currentMission.sprayLevelMaxValue, g_currentMission.sprayLevelMaxValue, g_currentMission.sprayLevelFirstChannel, g_currentMission.sprayLevelNumChannels)
        --end
    end
    spec.supportedSprayTypes = {}
    spec.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#fillUnitIndex"), 1)
	spec.sprayEffectAllowed = true 
	if self:getFillUnitExists(spec.fillUnitIndex) == false then
		spec.sprayEffectAllowed = false
		spec.fillUnitIndex = 1
	end
    spec.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.sprayer#unloadInfoIndex"), 1)
    spec.fillVolumeIndex = getXMLInt(self.xmlFile, "vehicle.sprayer#fillVolumeIndex")
    spec.dischargeUVScrollSpeed = { StringUtil.getVectorFromString(Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.sprayer#fillVolumeDischargeScrollSpeed"), "0 0 0")) }
    spec.needsToBeFilledToTurnOn = false
    spec.useSpeedLimit = false
    spec.isWorking = false
    spec.workAreaParameters = {}
    spec.workAreaParameters.sprayVehicle = nil
    spec.workAreaParameters.sprayVehicleFillUnitIndex = nil
    spec.workAreaParameters.lastChangedArea = 0
    spec.workAreaParameters.lastTotalArea = 0
end


function sprayerEffect:onPreDetach(attacherVehicle, jointDescIndex)
    if attacherVehicle.setIsTurnedOn ~= nil and attacherVehicle:getIsTurnedOn() then
        attacherVehicle:setIsTurnedOn(false)
    end
end


function sprayerEffect:onSetLowered(isLowered)
    local spec = self.spec_sprayerEffect
    if spec.activateOnLowering then
        self:setIsTurnedOn(isLowered, true)
    end

	
	
	
end


function sprayerEffect:onSprayTypeChange(activeSprayType)
    local spec = self.spec_sprayerEffect
    for _, sprayType in ipairs(spec.sprayTypes) do
        ObjectChangeUtil.setObjectChanges(sprayType.objectChanges, sprayType == activeSprayType)
    end
end


function sprayerEffect:onStartWorkAreaProcessing(dt)
    local spec = self.spec_sprayerEffect
    local sprayVehicle = nil
    local sprayVehicleFillUnitIndex = nil
    local fillType = self:getFillUnitFillType(self:getSprayerFillUnitIndex())
    local usage = self:getSprayerUsage(fillType, dt)
    local sprayFillLevel = self:getFillUnitFillLevel(self:getSprayerFillUnitIndex())
    if sprayFillLevel > 0 then
        sprayVehicle = self
        sprayVehicleFillUnitIndex = self:getSprayerFillUnitIndex()
    else
        for _, supportedSprayType in ipairs(spec.supportedSprayTypes) do
            for _, src in ipairs(spec.fillTypeSources[supportedSprayType]) do
                local vehicle = src.vehicle
                if vehicle:getIsFillUnitActive(src.fillUnitIndex) then
                    local vehicleFillType = vehicle:getFillUnitFillType(src.fillUnitIndex)
                    local vehicleFillLevel = vehicle:getFillUnitFillLevel(src.fillUnitIndex)
                    if vehicleFillLevel > 0 and vehicleFillType == supportedSprayType then
                        sprayVehicle = vehicle
                        sprayVehicleFillUnitIndex = src.fillUnitIndex
                        fillType = sprayVehicle:getFillUnitFillType(sprayVehicleFillUnitIndex)
                        usage = self:getSprayerUsage(fillType, dt)
                        sprayFillLevel = vehicleFillLevel
                        break
                    end
                else
                    -- if the ai is active we try to activate the source to activate the fill unit
                        if vehicle.setIsTurnedOn ~= nil then
                            if not vehicle:getIsTurnedOn() then
                                vehicle:setIsTurnedOn(true)
                            end
                        end
                end
            end
        end
    end
    spec.workAreaParameters.sprayType = g_sprayTypeManager:getSprayTypeIndexByFillTypeIndex(fillType)
    spec.workAreaParameters.sprayFillType = fillType
    spec.workAreaParameters.sprayFillLevel = sprayFillLevel
    spec.workAreaParameters.usage = usage
    spec.workAreaParameters.sprayVehicle = sprayVehicle
    spec.workAreaParameters.sprayVehicleFillUnitIndex = sprayVehicleFillUnitIndex
    spec.workAreaParameters.lastChangedArea = 0
    spec.workAreaParameters.lastTotalArea = 0
    spec.workAreaParameters.lastStatsArea = 0
    spec.workAreaParameters.isActive = false
    spec.isWorking = false
end


function sprayerEffect:onStateChange(state, data)
    if state == Vehicle.STATE_CHANGE_ATTACH or state == Vehicle.STATE_CHANGE_DETACH or Vehicle.STATE_CHANGE_FILLTYPE_CHANGE then
        local spec = self.spec_sprayerEffect
        spec.fillTypeSources = {}
        local supportedFillTypes = self:getFillUnitSupportedFillTypes(self:getSprayerFillUnitIndex())
        spec.supportedSprayTypes = {}
        if supportedFillTypes ~= nil then
            for fillType, state in pairs(supportedFillTypes) do
                if state then
                    spec.fillTypeSources[fillType] = {}
                    table.insert(spec.supportedSprayTypes, fillType)
                end
            end
        end
        local root = self:getRootVehicle()
        FillUnit.addFillTypeSources(spec.fillTypeSources, root, self, spec.supportedSprayTypes)
    end
end


function sprayerEffect:onTurnedOff()
    local spec = self.spec_sprayerEffect
    if self.isClient then
        g_effectManager:stopEffects(spec.effects)
        g_animationManager:stopAnimations(spec.animationNodes)
        if spec.animationName ~= "" and self.stopAnimation ~= nil then
            self:stopAnimation(spec.animationName, true)
        end
        g_soundManager:stopSample(spec.samples.work)
        -- deactivate effects on all spray types (the spray type has may changed during activation)
        for _, sprayType in ipairs(spec.sprayTypes) do
            g_effectManager:stopEffects(sprayType.effects)
            g_animationManager:stopAnimations(sprayType.animationNodes)
            g_soundManager:stopSample(sprayType.samples.work)
            self:playAnimation(sprayType.turnedAnimation, -1, self:getAnimationTime(sprayType.turnedAnimation), true)
        end
    end
end


function sprayerEffect:onTurnedOn()
    local spec = self.spec_sprayerEffect
    if self.isClient then
        local sprayType = self:getActiveSprayType()
		if self:getFillUnitFillLevel(spec.fillUnitIndex) > 0 and spec.sprayEffectAllowed == true then													
			if self:getAreEffectsVisible() then
				local fillType = self:getFillUnitLastValidFillType(self:getSprayerFillUnitIndex())
				if fillType == FillType.UNKNOWN then
					fillType = self:getFillUnitFirstSupportedFillType(self:getSprayerFillUnitIndex())
				end
				g_effectManager:setFillType(spec.effects, fillType)
				g_effectManager:startEffects(spec.effects)
				if sprayType ~= nil then
					g_effectManager:setFillType(sprayType.effects, fillType)
					g_effectManager:startEffects(sprayType.effects)
					g_animationManager:startAnimations(sprayType.animationNodes)
				end
				g_animationManager:startAnimations(spec.animationNodes)
			end
		end
        if spec.animationName ~= "" and self.playAnimation ~= nil then
            self:playAnimation(spec.animationName, 1, self:getAnimationTime(spec.animationName), true)
        end
        g_soundManager:playSample(spec.samples.work)
        if sprayType ~= nil then
            g_soundManager:playSample(sprayType.samples.work)
            self:playAnimation(sprayType.turnedAnimation, 1, self:getAnimationTime(sprayType.turnedAnimation), true)
        end
    end
end


function sprayerEffect:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local activeSprayType = self:getActiveSprayType()
    if activeSprayType ~= nil then
        local spec = self.spec_sprayerEffect
        if activeSprayType ~= spec.lastActiveSprayType then
            SpecializationUtil.raiseEvent(self, "onSprayTypeChange", activeSprayType)
            spec.lastActiveSprayType = activeSprayType
        end
    end
end


function sprayerEffect:onUpdate(dt)
	local spec = self.spec_sprayerEffect
	local fillLevel = self:getFillUnitFillLevel(spec.fillUnitIndex)
	if self.isClient then
		if fillLevel == 0 then
			g_effectManager:stopEffects(spec.effects)
			g_animationManager:stopAnimations(spec.animationNodes)
			if spec.animationName ~= "" and self.stopAnimation ~= nil then
				self:stopAnimation(spec.animationName, true)
			end
			g_soundManager:stopSample(spec.samples.work)
			-- deactivate effects on all spray types (the spray type has may changed during activation)
			for _, sprayType in ipairs(spec.sprayTypes) do
				g_effectManager:stopEffects(sprayType.effects)
				g_animationManager:stopAnimations(sprayType.animationNodes)
				g_soundManager:stopSample(sprayType.samples.work)
				self:playAnimation(sprayType.turnedAnimation, -1, self:getAnimationTime(sprayType.turnedAnimation), true)
			end
		end
    end


	if self:getIsLowered() == false then
		g_effectManager:stopEffects(spec.effects)
		g_animationManager:stopAnimations(spec.animationNodes)
		if spec.animationName ~= "" and self.stopAnimation ~= nil then
			self:stopAnimation(spec.animationName, true)
		end
		g_soundManager:stopSample(spec.samples.work)
		-- deactivate effects on all spray types (the spray type has may changed during activation)
		for _, sprayType in ipairs(spec.sprayTypes) do
			g_effectManager:stopEffects(sprayType.effects)
			g_animationManager:stopAnimations(sprayType.animationNodes)
			g_soundManager:stopSample(sprayType.samples.work)
			self:playAnimation(sprayType.turnedAnimation, -1, self:getAnimationTime(sprayType.turnedAnimation), true)
		end
	end
	
	if self:getIsLowered() == true then
		if self:getIsTurnedOn() == true then
			if self:getFillUnitFillLevel(spec.fillUnitIndex) > 0 and spec.sprayEffectAllowed == true then													
				if self:getAreEffectsVisible() then
					local fillType = self:getFillUnitLastValidFillType(self:getSprayerFillUnitIndex())
					if fillType == FillType.UNKNOWN then
						fillType = self:getFillUnitFirstSupportedFillType(self:getSprayerFillUnitIndex())
					end
					g_effectManager:setFillType(spec.effects, fillType)
					g_effectManager:startEffects(spec.effects)
					if sprayType ~= nil then
						g_effectManager:setFillType(sprayType.effects, fillType)
						g_effectManager:startEffects(sprayType.effects)
						g_animationManager:startAnimations(sprayType.animationNodes)
					end
					g_animationManager:startAnimations(spec.animationNodes)
				end
			end
		end
	end




end





function sprayerEffect.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(FillUnit, specializations) and
           SpecializationUtil.hasSpecialization(WorkArea, specializations) and
           SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations)
end


function sprayerEffect:processSprayerArea(workArea, dt)
    local spec = self.spec_sprayerEffect
    if spec.workAreaParameters.sprayFillLevel <= 0 then
        return 0, 0
    end
    local sx,_,sz = getWorldTranslation(workArea.start)
    local wx,_,wz = getWorldTranslation(workArea.width)
    local hx,_,hz = getWorldTranslation(workArea.height)
    local changedArea, totalArea = FSDensityMapUtil.updateSprayArea(sx,sz, wx,wz, hx,hz, spec.workAreaParameters.sprayType)
    spec.workAreaParameters.isActive = true
    spec.workAreaParameters.lastChangedArea = spec.workAreaParameters.lastChangedArea + changedArea
    spec.workAreaParameters.lastStatsArea = spec.workAreaParameters.lastStatsArea + changedArea
    spec.workAreaParameters.lastTotalArea = spec.workAreaParameters.lastTotalArea + totalArea
    if self:getLastSpeed() > 1 then
        spec.isWorking = true
    end
    return changedArea, totalArea
end


